/*
* Maciej Lech, grupa F1ISIII
* 
* Praca domowa z przedmiotu ZPR: Opis i przykad uycia Boost Static Assert.
*
* Przykad zakada korzystanie z bibliotek Boost w wersji 1.38
*
* UWAGA:
* Jako, e ten przykad ma pokazywa makro dziaajce w czasie kompilacji, oraz
* powodujce bdy kompilacji, zawiera on fragmenty kompilowane warunkowo.
* Gdy zdefiniujemy (za pomoc -Dmakro dla gcc, bd /Dmakro dla cl.exe) odpowiednie
* makro, spowodujemy bd kompilacji opisany w odpowiadajcym mu fragmencie tego
* przykadu. Ma to na celu zapoznanie z komunikatami kompilatora, jakie ten wypisuje,
* gdy Boost Assert otrzyma warto faszyw (ale te inne bdy wynikajce, np. z bdnego
* uycia Boost Assert).
* Oczywicie, odsyam po inne przykady do dokumentacji makra:
* http://www.boost.org/doc/libs/1_38_0/doc/html/boost_staticassert.html
*/

#include <boost/static_assert.hpp>
#include <cassert>
#include <limits>
#include <iostream>

/*
	* Biblioteka Boost Static Assert udostpnia pojedyncze
	* makro o nazwie BOOST_STATIC_ASSERT. Makro to jest
	* analogiczne do popularnego makra assert ze standardowej
	* biblioteki jzyka C z t jednak podstawow rnic,
	* e dziaa w czasie kompilacji kodu, a nie jego wykonania.
	* W przypadku, gdy warto wyraenia podanego jako argument
	* makra jest faszem, BOOST_STATIC_ASSERT powoduje bd
	* kompilacji. Naley zaznaczy, e warto argumentu makra
	* musi da si obliczy w trakcie kompilacji, eby makro
	* zadziao. W zwizku z tym ponisza konstrukcja zakoczy si
	* bdem kompilacji (i to bynajmniej nie wywoanym intencjonalnie
	* przez BOOST_STATIC_ASSERT):

	//...
*/

void funcWithI() {

	int i = 5;

	// rb co z i by moe zmieniajc jego warto...

#ifdef BOOST_STATIC_ASSERT_ERROR1
	BOOST_STATIC_ASSERT(i > 0); // bd - warto i nieznana podczas kompilacji
#else
	assert(i > 0); // OK
#endif
}

/*
	//...

	* W powyszym kodzie powinno zosta uyte makro assert, gdy ewidentnie
	* chcemy sprawdzi warto dostpn jedynie w czasie dziaania programu.
	* Aby wywoa bd kompilacji tego przykadu wynikajcy z prby uycia makra
	* BOOST_STATIC_ASSERT dla wyraenia nie dajcego si obliczy w czasie
	* kompilacji naley skompilowa ten plik ze zdefiniowanym makrem
	* BOOST_STATIC_ASSERT_ERROR1.
	*
	* Makro BOOST_STATIC_ASSERT mona umieci wszdzie tam, gdzie ,mona
	* umiescza deklaracje, a wic w przestrzeniach nazw, w klasach, funkcjach.
	*
	* *************************************************************************
	* BOOST_STATIC_ASSERT w przestrzeniach nazw
	* *************************************************************************
	* BOOST_STATIC_ASSERT w zasigu przestrzeni nazw jest uyteczne, gdy chcemy
	* sprawdzi, czy platforma, na ktrej dokonywana jest kompilacja naszego
	* programu spenia okrelone wymagania. Poniej chcemy zagwarantowa, e
	* platforma, dla ktrej kompilujemy kod ma typ long long int dugoci co najmniej
	* 32 bitw.
*/

namespace boost_static_assert_space {

	BOOST_STATIC_ASSERT(sizeof(unsigned long int) >= 4);
}

/*
	*
	* *************************************************************************
	* BOOST_STATIC_ASSERT w zasigu funkcji
	* *************************************************************************
	* Uycie makra z biblioteki Boost jest rwnie (a moe nawet bardziej)
	* przydatne w zasigu generycznych funkcji, do sprawdzania parametrw szablonu.
	* Poniej przykad funkcji liczcej silni, ktra akceptuje jedynie liczby
	* cakowite bez znaku (dla innych silnia jest niezdefiniowana):
*/

template<typename T> T factorial(T arg) {
	BOOST_STATIC_ASSERT(std::numeric_limits<T>::is_integer && !std::numeric_limits<T>::is_signed);
	
	T result = T(1);
	if (arg > T(0))
		while (arg > T(0)) {
			result *= arg;
			--arg;
		}
	return result;
}

/*
	* Uwaga do powyszego przykadu: Oczywicie moemy np. zaimplementowa klas, ktra
	* bdzie "udawa" liczb cakowit bez znaku o nieograniczonej iloci bitw
	* (przeciajc odpowiednie operatory) i wyspecjalizowa klas std::numeric_limits,
	* by poprzez is_integer zwracaa true, a przez is_signed false dla naszego nowego typu.
	* To pozwoli uywa powyszej funkcji dla naszych "spreparowanych" liczb.
	*
	* *************************************************************************
	* BOOST_STATIC_ASSERT w zasigu klasy
	* *************************************************************************
	* Makra BOOST_STATIC_ASSERT mona uywa rwnie w zasigu klasy. Najbardziej przydaje
	* si to znw w przypadku szablonw, gdy pozwala upewni si, e argument szablonu
	* spenia stawiane przed nim wymagania, ktrych inaczej nie daoby si sprawdzi podczas
	* kompilacji. Poniej znajduje si przykad klasy implementujcej tekstur dla potrzeb
	* grafiki, gdzie uyto bardzo wydajnych algorytmw, ktre maj jednak to ograniczenie,
	* e dziaaj tylko gdy rozmiary tekstury s potg liczby 2:
*/

template<unsigned int SIZE> class SquareOldStyleTexture {
	BOOST_STATIC_ASSERT((SIZE & (SIZE - 1)) == 0);
public:
	inline unsigned int getSize() { return SIZE; }
};

/*
	* **************************************************************************
	* Uwagi kocowe
	* **************************************************************************
	* Makro BOOST_STATIC_ASSERT pozwala wychwyci bdne (nizamierzone) uycia
	* szablonw, bd mechanizmw zanim objawi si awari dziaajcej aplikacji.
	* Ma jednak przynajmniej dwie wady:
	* * nie mona przekaza adnego komunikatu dotyczcego bdu asercji.
	* * bd asercji pojawia si w linii, w ktrej uyto makra, a nie w linii
	*   w ktrej wystpuje deklaracja, bd wywoanie, ktre spowodowao ten bd.
	* 
	* Powysze dwie rzeczy sprawiaj, e BOOST_STATIC_ASSERT jest dosy niewygodny
	* w uyciu.
	*
	* **************************************************************************
	* Na koniec przykadowy program.
	* **************************************************************************
	* Poniszy program mona skompilowa z makrami BOOST_STATIC_ASSERT_ERROR2
	* oraz BOOST_STATIC_ASSERT_ERROR3, ktre spowoduj bd kompilacji pokazujc
	* prawidowe dziaanie makra BOOST_STATIC_ASSERT.
*/
int main() {

#ifdef BOOST_STATIC_ASSERT_ERROR2
	int n = 5; //funckja factorial nie skompiluje si dla typu int, ktry jest typem ze znakiem
#else
	unsigned long int n = 5; //tu jest OK
#endif

	std::cout << n << "! == " << factorial(n) << std::endl;

#ifdef BOOST_STATIC_ASSERT_ERROR3
	SquareOldStyleTexture<432> texture; //ta linia nie skompiluje si, gdy 432 nie jest potg liczby 2.
#else
	SquareOldStyleTexture<512> texture; // tekstura o poprawnym rozmiarze
#endif

	std::cout << "Tekstura ma rozmiar " << texture.getSize() << " pikseli." << std::endl;

	return 0;
}
